#!/usr/bin/python3

import argparse
import pdb
import sys

predefined = {
    "R0": 0,
    "R1": 1,
    "R2": 2,
    "R3": 3,
    "R4": 4,
    "R5": 5,
    "R6": 6,
    "R7": 7,
    "R8": 8,
    "R9": 9,
    "R10": 10,
    "R11": 11,
    "R12": 12,
    "R13": 13,
    "R14": 14,
    "R15": 15,
    "SCREEN": 16384,
    "KBD": 24576,
    "SP": 0,
    "LCL": 1,
    "ARG": 2,
    "THIS": 3,
    "THAT": 4
}

dst_table = {
    "null": "000",
    "M": "001",
    "D": "010",
    "MD": "011",
    "A": "100",
    "AM": "101",
    "AD": "110",
    "AMD": "111",
}

cmp_table = {
    "0":   "0101010",
    "1":   "0111111",
    "-1":  "0111010",
    "D":   "0001100",
    "A":   "0110000",
    "!D":  "0001101",
    "!A":  "0110001",
    "-D":  "0001111",
    "-A":  "0110011",
    "D+1": "0011111",
    "A+1": "0110111",
    "D-1": "0001110",
    "A-1": "0110010",
    "D+A": "0000010",
    "D-A": "0010011",
    "A-D": "0000111",
    "D&A": "0000000",
    "D|A": "0010101",
    "M":   "1110000",
    "!M":  "1110001",
    "-M":  "1110011",
    "M+1": "1110111",
    "M-1": "1110010",
    "D+M": "1000010",
    "D-M": "1010011",
    "M-D": "1000111",
    "D&M": "1000000",
    "D|M": "1010101",
}

jmp_table = {
    "null": "000",
    "JGT": "001",
    "JEQ": "010",
    "JGE": "011",
    "JLT": "100",
    "JNE": "101",
    "JLE": "110",
    "JMP": "111",
}

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("filename")
    args = parser.parse_args()

    with open(args.filename) as f:
        data = f.read()

    no_comments = strip_comments(data)
    no_whitespace = strip_whitespace(no_comments)
    symbol = symbols(no_whitespace)
    sys.stdout.write(instructions(symbol))

def strip_comments(data):
    ret = ""
    in_comment = False

    for c in data:
        if c == "/":
            in_comment = True
        elif c == "\n":
            in_comment = False

        if not in_comment:
            ret += c

    return ret

def strip_whitespace(data):
    ret = ""
    prev = "\n"

    for c in data:
        if c == "\n" and prev == "\n":
            pass
        elif c == " " or c == "\t":
            pass
        else:
            ret += c
            prev = c

    return ret

def symbols(data):
    i = 0
    ret = ""
    line_count = 0

    while i < len(data):
        if data[i] == "(":
            i += 1
            temp = ""
            while data[i] != ")":
                temp += data[i]
                i += 1

            i += 1
            if not temp in predefined:
                predefined[temp] = line_count
        elif data[i] == "\n":
            line_count += 1
        i += 1

    i = 0
    variables = {}
    variable_count = 0
    while i < len(data):
        if data[i] == "@":
            i += 1
            temp = ""
            while data[i] != "\n":
                temp += data[i]
                i += 1

            if temp in predefined:
                ret += ("@" + str(predefined[temp]) + "\n")
            elif temp in variables:
                ret += ("@" + str(variables[temp]) + "\n")
            elif is_label(temp):
                ret += ("@" + str(16 + variable_count) + "\n")
                variables[temp] = 16 + variable_count
                variable_count += 1
            else:
                ret += ("@" + temp + "\n")
        elif data[i] == "(":
            i += 1
            temp = ""
            while data[i] != ")":
                temp += data[i]
                i += 1

            i += 1
        else:
            ret += data[i]
        i += 1

    return ret

def instructions(data):
    ret = ""
    i = 0
    dst_bits = "000"
    comp_bits = ""
    jump_bits = "000"

    while i < len(data):
        if data[i] == "@":
            i += 1
            ret += "0"
            temp = ""
            while data [i] != "\n":
                temp += data[i]
                i += 1

            ret += "{0:015b}".format(int(temp)) + "\n"
        elif data[i] != "\n":
            # we have a c-instruction
            ret += "111"
            instr = ""
            while data[i] != "\n":
                instr += data[i]
                i += 1

            # we have the instruction, questionable parsing logic follows
            j = 0
            if "=" in instr:
                # we need to pull the dest out
                dest = ""

                while instr[j] != "=":
                    dest += instr[j]
                    j += 1

                dst_bits = dst_table[dest]

                # pull the comp out
                comp = ""
                j += 1
                while j < len(instr) and instr[j] != ";":
                    comp += instr[j]
                    j += 1

                comp_bits = cmp_table[comp]
            else:
                # pull the comp out
                dst_bits = "000"
                comp = ""

                while instr[j] != ";":
                    comp += instr[j]
                    j += 1

                comp_bits = cmp_table[comp]

            if ";" in instr:
                j = 0
                while instr[j] != ";":
                    j += 1

                j += 1
                jump = ""
                while j < len(instr) and instr[j] != "\n":
                    jump += instr[j]
                    j += 1

                jump_bits = jmp_table[jump]
            else:
                jump_bits = "000"

            ret += comp_bits + dst_bits + jump_bits
            ret += "\n"

        i += 1
            
    return ret

def is_label(data):
    for c in data:
        if not (c >= "0" and c <= "9"):
            return True

    return False

if __name__ == "__main__":
    main()
